#include <SPI.h>
#include <SD.h>
#include <Ethernet.h>
#include <OneWire.h>

//ethernet shield
#define   REQ_BUF_SZ  60
byte mac[] = {
  0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED
};
IPAddress ip(192, 168, 2, 5);
EthernetServer server(80);
char HTTP_req[REQ_BUF_SZ] = {
  0
};
char req_index = 0;
String req_str;

File MainPage;

int MainStatus = 1;
//1 нормальная работа
//-1 ошибка чтерия SD карты
//-10 превышение температуры нв аварийном датчике

//OneWire
#define DS1820_PIN 23 //Пин подключения датчиков 
#define MAX_DS1820 5 //максимальное количество датчиков
byte ds1820cnt = 0; //количество найденных датчиков температуры
byte ds1820_rom[MAX_DS1820][8]; //массив адресов датчиков
byte ds1820_data[MAX_DS1820][12]; //массив данных датчиков
float dsTemp[MAX_DS1820]; //массив с температурой
OneWire ds1820Wire(DS1820_PIN);
long timerTemp = 0;
long timerTempInterval = 1000;

#define ALARM_NUM  2 //номер аварийного датчика температуры -1
#define ALARM_TEMP 55 //при превышении температуры РМЦ переводится в пежим паузы и закрываются клапана.

//старт-стоп
boolean StartStopEnable = false;
boolean StartStopSrab = false;
float StartStopTemp = 0;
#define STARTSTOP_PIN 39 //клапан ХД-АКР

//управление клапанами с ШИМ
#define VL_CNT 4//количество клапанов
byte VL_PIN[VL_CNT];//пин подключения
byte VLPWM[VL_CNT];//% ШИМ
int VLCnt[VL_CNT];
int VLCntM[VL_CNT];
byte VLMode[VL_CNT]; //0 - закрыт 1 - шим регулирование 2 - открыт
byte VL_INVERSE[VL_CNT];//инверсное управление задается в void setup ()

//управление реле
#define REL_CNT 4 //количество реле
byte REL_PIN[REL_CNT]; //пин подключения реле
byte REL_STATE[REL_CNT]; //состояние реле
boolean REL_INVERSE[REL_CNT];//инверсное управление задается в void setup ()

//датчик давления
#define MPX_PIN A9
int U_MPX = 0;
int Pressure;

//РМЦ-3500
#define RMC_PAUSE_PIN 25 //пауза
#define RMC_PWM_PIN 27  //шим
#define TEN_POWER 3000 //Мощность тэн-а пока нишде не используется
int RMCPowerMode = 0;
int RMCPower = 0;
int RMCCnt = 0;

int motorenable = 0;
int motordir = 0;
int rpm = 0;


//Таймер 50Гц для РМЦ-3500 и клапана отбора
void SetupTimer1() {
  TCCR1B = (1 << WGM12);
  TCCR1B |= (1 << CS10);
  TCCR1B |= (1 << CS12);
  OCR1A = 1249;
  TIMSK1 = (1 << OCIE1A);
}

ISR(TIMER1_COMPA_vect) {
  //РМЦ-3500
  if (RMCPowerMode == 1) {
    if (RMCCnt == 99) {
      digitalWrite(RMC_PWM_PIN, HIGH);
      RMCCnt = 0;
    }
    if (RMCCnt == RMCPower) {
      digitalWrite(RMC_PWM_PIN, LOW);
    }
    RMCCnt++;
  }
  //шим на клапанах
  for (int i = 0; i < VL_CNT; i++) {
  if (VLMode[i] == 1) {
    if (StartStopEnable && StartStopSrab) {
      digitalWrite(VL_PIN[i], LOW - VL_INVERSE[i]);
    }
    else {
      if (VLCnt[i] == 99) {
        digitalWrite(VL_PIN[i], HIGH - VL_INVERSE[i]);
        VLCnt[i]= 0;
        VLCntM[i] = 0;
      }
      if (VLCnt[i] == VLPWM[i]) {
        digitalWrite(VL_PIN[i], LOW - VL_INVERSE[i]);
      }
      VLCntM[i] ++;
      if (VLCntM[i] == 5) {
        VLCnt[i]++;
        VLCntM[i] = 0;
      }
    }
  }
  }
}

void connectToSensors() {
  int i, j;
  for (i = 0; i == (MAX_DS1820-1); i++) {
    dsTemp[i] = 0;
  }
  while (ds1820Wire.search(ds1820_rom[ds1820cnt])) {
    if (OneWire::crc8(ds1820_rom[ds1820cnt], 7) == ds1820_rom[ds1820cnt][7]) {
      ds1820cnt++;
    }
    else {
      for (i = 0; i < 8; i++) {
        Serial.print(ds1820_rom[ds1820cnt][i], HEX);
        Serial.print(":");
      }
      Serial.println(" Read ROM CRC Error ");
    }
    if (ds1820cnt >= MAX_DS1820) break;
  }
  ds1820Wire.reset_search();
  delay(250);
  Serial.print("DS18x20 count=");
  Serial.println(ds1820cnt);
  for (i = 0; i < ds1820cnt; i++) {
    Serial.print(i + 1);
    Serial.print(": ");
    for (j = 0; j < 8; j++) {
      Serial.print(ds1820_rom[i][j], HEX);
      Serial.print(":");
    }
    switch (ds1820_rom[i][0]) {
      case 0x10:
        Serial.print(" chip 18S20");
        break;
      case 0x22:
        Serial.print(" chip 1822");
        break;
      case 0x28:
        Serial.print(" chip 18B20");
        break;
      default:
        Serial.print(" Device is not a DS18x20 family device.");
        return;
    }
    Serial.println();
  }
  ds1820Wire.reset();
  ds1820Wire.write(0xCC);
  ds1820Wire.write(0x44, 1);
  delay(1000);
}

void getTemperature () {
  for (int i = 0; i < ds1820cnt-1; i++) {
    ds1820Wire.reset();
    ds1820Wire.select(ds1820_rom[i]);
    ds1820Wire.write(0xBE);
    for (byte j = 0; j < 9; j++) {
      ds1820_data[i][j] = ds1820Wire.read();
    }
    if (OneWire::crc8(ds1820_data[i], 8) == ds1820_data[i][8]) {
      int16_t raw = (ds1820_data[i][1] << 8) | ds1820_data[i][0];
      float celsius = (float)raw / 16.0;
      dsTemp[i] = celsius;
    }
    else {
      Serial.println("Read CRC error");
    }
  }
  ds1820Wire.reset();
  ds1820Wire.write(0xCC);
  ds1820Wire.write(0x44, 1);
  timerTemp = millis();
}

void SetParms(void)
{
  int len = 0;
  String t = "";
  if (req_str.indexOf("StartStop=1") >= 0) {
    StartStopEnable = true;
    Serial.println("start-stop on");
  }
  if (req_str.indexOf("StartStop=0") >= 0){
     StartStopEnable = false;
     Serial.println("start-stop off");
  }
  if (req_str.indexOf("StartStopTemp=") >= 0) {
    t = req_str.substring(req_str.indexOf("StartStopTemp=") + 14, req_str.length());
    len = t.indexOf("&");
    if (len == 0) len = t.length();
    t = t.substring(0, len);
    char buf[t.length() + 1];
    t.toCharArray(buf, sizeof(buf));
    StartStopTemp = atof(buf);
    Serial.print("start-stop temp ");
    Serial.println(StartStopTemp);
  }
  for (int i=0; i < VL_CNT; i++) {
    if (req_str.indexOf("Valve"+String(i,DEC)+"Mode=0") >= 0) {
      VLMode[i]=0;
      digitalWrite(VL_PIN[i], LOW - VL_INVERSE[i]);
      Serial.println("Valve"+String(i,DEC)+" on");
    }
    if (req_str.indexOf("Valve"+String(i,DEC)+"Mode=1") >= 0 ) {
      VLMode[i] = 1;
      VLPWM[i] = 0;
      if (req_str.indexOf("Valve"+String(i,DEC)+"Val=") >= 0) {
        t = req_str.substring(req_str.indexOf("Valve"+String(i,DEC)+"Val=") + 10, req_str.length());
        len = t.indexOf("&");
        if (len == 0) len = t.length();
        t = t.substring(0, len);
        VLPWM[i] = t.toInt();
        if (VLPWM[i] < 0) VLPWM[i] = 0;
        if (VLPWM[i] > 100) VLPWM[i] = 100;
      }
      digitalWrite(VL_PIN[i], LOW - VL_INVERSE[i]);
      VLCnt[i] = 0;
      VLCntM[i] = 0;
      Serial.print("Valve"+String(i,DEC)+" PWM: ");
      Serial.println(VLPWM[i]);
    }
    if (req_str.indexOf("Valve"+String(i,DEC)+"Mode=2") >= 0 ) {
      VLMode[i] = 2;
      digitalWrite(VL_PIN[i], HIGH - VL_INVERSE[i]);
      Serial.println("Valve"+String(i,DEC)+" off");
    }
  }
  for (int i=0; i < REL_CNT; i++) {
    if (req_str.indexOf("Relay"+String(i,DEC)+"=0") >= 0) {
      REL_STATE[i] = 0;
      digitalWrite(REL_PIN[i],LOW - REL_INVERSE[i]);
    }
    if (req_str.indexOf("Relay"+String(i,DEC)+"=1") >= 0) {
      REL_STATE[i] = 1;
      digitalWrite(REL_PIN[i],HIGH - REL_INVERSE[i]);
    }
  }
  if (req_str.indexOf("PowerMode=0") >= 0 ) {
    //пауза
    RMCPowerMode = 0;
    digitalWrite(RMC_PWM_PIN, LOW);
    digitalWrite(RMC_PAUSE_PIN, HIGH);
    Serial.println("RMC pause");
  }
  if (req_str.indexOf("PowerMode=1") >= 0 ) {
    //шим регулирование
    RMCPowerMode = 1;
    RMCPower = 0;
    if (req_str.indexOf("PowerVal=") >= 0) {
      t = req_str.substring(req_str.indexOf("PowerVal=") + 9, req_str.length());
      len = t.indexOf("&");
      if (len == 0) len = t.length();
      t = t.substring(0, len);
      RMCPower = t.toInt();
      if (RMCPower < 0) RMCPower = 0;
      if (RMCPower > 100) RMCPower = 100;
    }
    digitalWrite(RMC_PAUSE_PIN, LOW);
    digitalWrite(RMC_PWM_PIN, HIGH);
    Serial.print("RMC PWM ");
    Serial.println(RMCPower);
    RMCCnt = 0;
  }
  if (req_str.indexOf("PowerMode=2") >= 0 ) {
    //разгон
    RMCPowerMode = 2;
    digitalWrite(RMC_PAUSE_PIN, LOW);
    digitalWrite(RMC_PWM_PIN, HIGH);
    Serial.println("RMC razgon");
  }
    if (req_str.indexOf("PumpMode=1") >= 0) {
      motorenable=true;
//      digitalWrite(MOTORENABLE_IO,LOW);
    }
    if (req_str.indexOf("PumpMode=0") >=0 ) {
      motorenable=false;
//      digitalWrite(MOTORENABLE_IO,HIGH);
    }
    if (req_str.indexOf("PumpRevers=1") >=0 ) {
      motordir = !motordir;
    }
    if (motordir) {
//      digitalWrite(MOTORDIR_IO,LOW);
    }
    else {
//      digitalWrite(MOTORDIR_IO,HIGH);
    }
    if (req_str.indexOf("PumpRPM=") >=0 ){
      t=req_str.substring(req_str.indexOf("PumpRPM=")+8,req_str.length());
      len = t.indexOf("&");
      if (len == 0) len = t.length();
      t = t.substring(0,len);
      rpm = t.toInt();
      if (rpm < 0) {
        rpm = 0;
      }
 //     if (rpm > MAX_RPM) {
//        rpm = MAX_RPM;
 //     }
  //    FrequencyTimer2::setPeriod(1000000/((rpm/2)*((MOTOR_STEP*MICROSTEP)/60)));
    }
}

// send the XML file
void XML_response(EthernetClient cl)
{
  int i;
  cl.println("<?xml version = \"1.0\" ?>");
  cl.println("<inputs>");
  cl.print("<MainStatus>");
  cl.print(MainStatus);
  cl.println("</MainStatus>");
  for (int i = 0; i < MAX_DS1820; i++) {
    cl.print("<dstemp");
    cl.print(i);
    cl.print(">");
    cl.print(dsTemp[i]);
    cl.print("</dstemp");
    cl.print(i);
    cl.println(">");
  }
  cl.print("<pressure0>");
  cl.print(Pressure);
  cl.println("</pressure0>");
  for (int i = 0; i <= 2; i++) {
    cl.print("<analog");
    cl.print(i);
    cl.print(">");
    cl.print("0");
    cl.print("</analog");
    cl.print(i);
    cl.println(">");
  }
  cl.print("<StartStopEn>");
  if (StartStopEnable) {
    cl.print(1);
  }
  else {
    cl.print(0);
  }
  cl.println("</StartStopEn>");

  cl.print("<StartStopSrab>");
  if (StartStopSrab) {
    cl.print(1);
  }
  else {
    cl.print(0);
  }
  cl.println("</StartStopSrab>");
  cl.print("<StartStopTemp>");
  cl.print(StartStopTemp);
  cl.println("</StartStopTemp>");
  cl.println("<TKubStopEn>0</TKubStopEn>");
  cl.println("<TKubStopTemp>200</TKubStopTemp>");
  cl.println("<TKubStopStatus>0</TKubStopStatus>");
  cl.println("<AutoOtborEn>0</AutoOtborEn>");
  cl.println("<AutoOtborTemp>0</AutoOtborTemp>");  
  cl.print("<RMCPowerMode>");
  cl.print(RMCPowerMode);
  cl.println("</RMCPowerMode>");
  cl.print("<RMCPower>");
  cl.print(RMCPower);
  cl.println("</RMCPower>");
  cl.print("<PumpMode>");
    if (motorenable) {
      cl.print(1);
    }
    else {
      cl.print(0);
    }
  cl.println("</PumpMode>");
  cl.print("<PumpDir>");
    if (motordir) {
      cl.print(1);
    }
    else {
      cl.print(0);
    }
  cl.println("</PumpDir>");
  cl.print("<PumpRPM>");
  cl.print(rpm);
  cl.println("</PumpRPM>");
  for (int i = 0; i < VL_CNT; i++) {
    cl.print("<Valve");
    cl.print(i);
    cl.print("Mode>");
    cl.print(VLMode[i]);
    cl.print("</Valve");
    cl.print(i);
    cl.println("Mode>");
    cl.print("<Valve");
    cl.print(i);
    cl.print("PWM>");
    cl.print(VLPWM[i]);
    cl.print("</Valve");
    cl.print(i);
    cl.println("PWM>");
  }
  for (int i = 0; i < REL_CNT; i++) {
    cl.print("<Relay");
    cl.print(i);
    cl.print(">");
    cl.print(REL_STATE[i]);
    cl.print("</Relay");
    cl.print(i);
    cl.println(">");
  }
  cl.println("</inputs>");
}

void StrClear(char *str, char length)
{
  for (int i = 0; i < length; i++) {
    str[i] = 0;
  }
}

char StrContains(char *str, char *sfind)
{
  char found = 0;
  char index = 0;
  char len;
  len = strlen(str);
  if (strlen(sfind) > len) {
    return 0;
  }
  while (index < len) {
    if (str[index] == sfind[found]) {
      found++;
      if (strlen(sfind) == found) {
        return 1;
      }
    }
    else {
      found = 0;
    }
    index++;
  }
  return 0;
}

void SendMainPage(EthernetClient cl) {
  if (MainStatus == -1) {    
    cl.println("<html><head><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" /><title>home distiller arduino web controller</title></head>");
    cl.println("<body><h1>Ошибка чтения SD карты</h1></body>");
    return;
  }
  MainPage = SD.open("index.htm");
    if (MainPage) {
      while(MainPage.available()) {
        cl.write(MainPage.read());
      }
      MainPage.close();
  }
}

void setup () {
  Serial.begin(9600);
  //назначаем пины управления клапанами
  VL_PIN[0]=22;
  VL_PIN[1]=24;
  VL_PIN[2]=26;
  VL_PIN[3]=28;
  // и реле
  REL_PIN[0]=30;
  REL_PIN[1]=32;
  REL_PIN[2]=34;
  REL_PIN[3]=36;
  //инверсное управление
  VL_INVERSE[0]=false;
  VL_INVERSE[1]=false;
  VL_INVERSE[2]=false;
  VL_INVERSE[3]=false;
  REL_INVERSE[0]=false;
  REL_INVERSE[1]=false;
  REL_INVERSE[2]=false;
  REL_INVERSE[3]=false;
  for (int i = 0; i < VL_CNT; i++) {
    VLMode[i]=0;
    pinMode(VL_PIN[i], OUTPUT);
    digitalWrite(VL_PIN[i], LOW - VL_INVERSE[i]);
  }
  for (int i = 0; i < REL_CNT; i++) {
    REL_STATE[i]=0;
    pinMode(REL_PIN[i], OUTPUT);
    digitalWrite(REL_PIN[i], LOW - REL_INVERSE[i]);
  }
  pinMode(STARTSTOP_PIN, OUTPUT);
  pinMode(RMC_PAUSE_PIN, OUTPUT);
  pinMode(RMC_PWM_PIN, OUTPUT);
  digitalWrite(RMC_PWM_PIN, LOW);
  digitalWrite(RMC_PAUSE_PIN, HIGH);
  connectToSensors();
  SetupTimer1();
  Ethernet.begin(mac, ip);
  server.begin();
  if (!SD.begin(4)) {
    Serial.println("initialization SD card failed!");
    MainStatus = -1;
    return;
  }
  Serial.println("initialization SD card done."); 
}

void loop () {
  //опрос датчиков
  if (millis() - timerTemp > timerTempInterval) {
    getTemperature();
    U_MPX = analogRead(MPX_PIN);
    Pressure = ((U_MPX * 4 - 160) / 5) / 10;
    timerTemp = millis();
    if ((dsTemp[1] >= StartStopTemp) && (StartStopEnable)){
      StartStopSrab = true;
    }
    else {
      StartStopSrab = false;
    }
  }
  if (dsTemp[ALARM_NUM] > ALARM_TEMP) {
    MainStatus = -10;
  }
  switch(MainStatus) {
    case -10: //авария по датчику температуры
    for (int i=0; i < VL_CNT; i++) {
      digitalWrite(VL_PIN[i],LOW - VL_INVERSE[i]);
      VLMode[i] = 0;
    }
    digitalWrite(RMC_PWM_PIN,LOW);
    digitalWrite(RMC_PAUSE_PIN,HIGH);
    RMCPowerMode = 0;
    RMCPower = 0;
    break;
  }
  EthernetClient client = server.available();
  if (client) {
    req_str = "";
    boolean currentLineIsBlank = true;
    boolean isFirstLine = true;
    while (client.connected()) {
      if (client.available()) {
        char c = client.read();
        if (req_index < (REQ_BUF_SZ - 1)) {
          HTTP_req[req_index] = c;
          req_index++;
          if (isFirstLine) {
            req_str += c;
          }
        }
        if (c == '\n' && currentLineIsBlank) {
          client.println("HTTP/1.1 200 OK");
          if (StrContains(HTTP_req, "ajaxinputs")) {
            client.println("Content-Type: text/xml");
            client.println("Connection: keep-alive");
            client.println();
            SetParms();
            XML_response(client);
          }
          else {
            client.println("Content-Type: text/html");
            client.println("Connection: keep-alive");
            client.println();
            SendMainPage(client);
          }
          req_index = 0;
          StrClear(HTTP_req, REQ_BUF_SZ);
          break;
        }
        if (c == '\n') {
          currentLineIsBlank = true;
          isFirstLine = false;
        }
        else if (c != '\r') {
          currentLineIsBlank = false;
        }
      } // end if (client.available())
    } // end while (client.connected())
    delay(1);
    client.stop();
  } // end if (client)
  }
